/* houghseq.c in houghpar.c */
/* Izracuna Houghov transform na zaporedni oziroma vzporedni nacin */
/* Janez Pers, 2001 */

#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <string.h>
#include "pgmio.c"
#include "hough.c"

#define MAXZNAKOV 255

main(int argc, char *argv[])
{
/* Spremenljivke za merjenje casa, ki ga program porabi za obdelavo */
time_t starttime;
time_t endtime;
double processingtime;

/* Imena datotek in kazalci na datoteke */
char infilename [MAXZNAKOV];
char outfilename [MAXZNAKOV];
FILE *infile;
FILE *outfile;

/* Vhodna in izhodna slika ter zacasni akumulator*/
int inimage[512][512];
int outimage[512][512];
float accumulator[512][512];

/* Splosnonamenski stevec in indeksa v sliko */
int i, x, y;

/* Upravljanje procesov in komunikacija med procesi*/
int process_n;
int fork_code;
pid_t process_table[10];
float *shmaddr;
int shmid;
int semid;
struct sembuf sbuf[2];

/* Preverimo stevilo vhodnih parametrov, ce argc!=3 izpisemo navodila */
if (argc!=3)
{
   printf("\nHOUGHX.C - Izracun Houghove transformacije. \n");
   printf("\nUporaba: HOUGHPSEQ ali HOUGHPAR <vhodna PGM slika> <izhodna PGM slika> \n");
   printf("Vhodna slika mora biti obvezno dimenzij 512x512. \n\n");
   exit (1);
};

/* Prenesemo vrednosti argumentov v ustrezni spremenljivki */
strcpy(infilename,  argv[1]);
strcpy(outfilename, argv[2]);

printf("\nHOUGHX.C - Izracun Houghove transformacije. \n");

/* Preberemo vhodno sliko */
pgmread(infilename, inimage);

/* Zapomni si cas zacetka racunanja */
starttime = time(NULL);
printf ("Zacetek racunanja ob: %s", ctime(&starttime));


/* Alociramo shared memory za izmenjavo rezultatov */
if( (shmid = shmget( IPC_PRIVATE, sizeof(float)*512*512, 0600 )) == -1 )
{
    printf("Napaka, shmget ni uspel\n");
    exit(0);
};

/* Alociramo semaforje za sinhronizacijo dostopa do shared memory */
if( (semid = semget(IPC_PRIVATE, 5, 0600)) == -1)
{
    printf("Napaka, semget ni uspel\n");
    exit(0);
};
/* Semaforje postavimo na 0 - otroci ne smejo dostopati do shared mem */
semctl(semid, 1, SETVAL, 0);
semctl(semid, 2, SETVAL, 0);
semctl(semid, 3, SETVAL, 0);
semctl(semid, 4, SETVAL, 0);

/* Naredimo 4 nove procese, ki bosta izvajala racunanje */
process_n = 0; /* roditelj ima process_n=0, otroci pa 1,2,3,... */

/* Prvi proces naredimo enostavno */
if (1)
{
   fork_code = fork();
   process_table[1] = fork_code;
   if (fork_code == 0) process_n = 1;
};

/* Pri drugem moramo preveriti, ali se nahajamo v roditelju */
if (fork_code > 0)
{
   fork_code = fork();
   process_table[2] = fork_code;
   if (fork_code == 0) process_n = 2;
};

/* Tretji proces... */
if (fork_code > 0)
{
   fork_code = fork();
   process_table[3] = fork_code;
   if (fork_code == 0) process_n = 3;
};

/* Cetrti proces... */
if (fork_code > 0)
{
   fork_code = fork();
   process_table[4] = fork_code;
   if (fork_code == 0) process_n = 4;
};

/* Roditelj izpise podatke o procesih */
if (process_n==0)
{
   for (i=1; i<5; i++)
   {
       printf ("process_n: %d; pid = %d \n",i, process_table[i]);
   };
   printf ("\n");
};

/* Vsak proces-otrok obdela eno cetrtino slike */
if (process_n==1) hough(inimage, accumulator, 0, 127);
if (process_n==2) hough(inimage, accumulator, 128, 255);
if (process_n==3) hough(inimage, accumulator, 256, 383);
if (process_n==4) hough(inimage, accumulator, 384, 511);
if (process_n>0)
{
   printf("Proces %d koncal racunanje, caka na semaforju. \n", process_n);
};

/* Roditelj in otroci se pripnejo na shared memory */
if ((shmaddr = shmat( shmid, 0, 0 )) == (void *) -1 )
{
    printf("Napaka, shmat ni uspel\n");
    exit (0);
};


/* Roditelj inicializira shared memory na 0 (262144=512*512) */
if (process_n==0)
{
for (i=0; i<262144; i++)
    {
       shmaddr[i]= 0;
    };
};

/* Procesi-otroci medtem cakajo na semaforju */
if (process_n>0)
{
    sbuf[0].sem_num = process_n;
    sbuf[0].sem_op = -1;
    sbuf[0].sem_flg = 0;
    semop (semid, sbuf, 1);
    printf("Proces %d dobil zelen semafor. \n", process_n);
};

/* Procesi-otroci pristejejo svoje podatke v shared memory */
if (process_n>0)
{
    for (y=0; y<512; y++)
    {
        for (x=0; x<512; x++)
        {
            shmaddr[y*512+x]= shmaddr[y*512+x] + accumulator[x][y];
        };
    };
};


/* Otroci sprostijo shared memory */
if (process_n>0)
{
    if(shmdt(shmaddr ) == -1)
    {
        printf("Napaka, shmdt za otroka ni uspel\n");
        exit(0);
    };
};

/* Procesi-otroci se koncajo tu */
if (process_n>0) exit(0);

/* Proces-roditelj najprej vsakemu od njih prizge zeleno na semaforju,
/* nato pa caka caka na vsakega od njih, da opravijo dostop do shared
/* memory in se koncajo. Od tu naprej se spet izvaja le roditelj */
for (i=1; i<5; i++)
{
   sbuf[0].sem_num = i;
   sbuf[0].sem_op = 1;
   sbuf[0].sem_flg = 0;
   semop (semid, sbuf, 1);
   waitpid(process_table[i],0,0);
};

/* Roditelj prepise podatke iz shared memory v svoj pomnilnik */
if (process_n==0)
{
    for (y=0; y<512; y++)
    {
        for (x=0; x<512; x++)
        {
            accumulator[x][y] = shmaddr[y*512+x];
        };
    };
};

/* Roditelj sprosti shared memory */
if (process_n==0)
{
    if(shmdt(shmaddr ) == -1)
    {
        printf("Napaka, shmdt za roditelja ni uspel\n");
        exit(0);
    };
};

/* Izbrisemo shared memory segment */
if( shmctl(shmid, IPC_RMID, 0 ) == -1)
{
    printf("Napaka, shmctl z SHM_RMID ni uspel\n");
};

/* Izbrisemo semafor */
if( semctl(semid, 0, IPC_RMID, 0 ) == -1)
{
    printf("Napaka, semctl z SHM_RMID ni uspel\n");
};

/* Zapomni si cas konca racunanja */
endtime = time(NULL);
printf ("Konec racunanja ob: %s \n", ctime(&endtime));

/* Izracunaj porabljen cas v sekundah */
processingtime = difftime(endtime, starttime);
printf ("Cas racunanja: %.0f sekund. \n", processingtime);

/* Skaliraj akumulator, ter ga prepisi v izhodno sliko */

acc2img(accumulator, outimage);

/* Zapisemo izhodno sliko */
pgmwrite(outfilename, outimage);

}
